@@ -47,11 +47,13 @@ public class ReaderView
 	private int               mXScroll;    // Scroll amounts recorded from events.
 	private int               mYScroll;    // and then accounted for in onLayout
 	private boolean           mReflow = false;
+	private boolean           mReflowChanged = false;
 	private final GestureDetector
 				  mGestureDetector;
 	private final ScaleGestureDetector
 				  mScaleGestureDetector;
 	private final Scroller    mScroller;
+	private final Stepper     mStepper;
 	private int               mScrollerLastX;
 	private int               mScrollerLastY;
 	private boolean           mScrollDisabled;
@@ -65,6 +67,7 @@ public class ReaderView
 		mGestureDetector = new GestureDetector(this);
 		mScaleGestureDetector = new ScaleGestureDetector(context, this);
 		mScroller        = new Scroller(context);
+		mStepper = new Stepper(this, this);
 	}
 
 	public ReaderView(Context context, AttributeSet attrs) {
@@ -72,6 +75,7 @@ public class ReaderView
 		mGestureDetector = new GestureDetector(this);
 		mScaleGestureDetector = new ScaleGestureDetector(context, this);
 		mScroller        = new Scroller(context);
+		mStepper = new Stepper(this, this);
 	}
 
 	public ReaderView(Context context, AttributeSet attrs, int defStyle) {
@@ -79,6 +83,7 @@ public class ReaderView
 		mGestureDetector = new GestureDetector(this);
 		mScaleGestureDetector = new ScaleGestureDetector(context, this);
 		mScroller        = new Scroller(context);
+		mStepper = new Stepper(this, this);
 	}
 
 	public int getDisplayedViewIndex() {
@@ -202,7 +207,7 @@ public class ReaderView
 		}
 		mScrollerLastX = mScrollerLastY = 0;
 		mScroller.startScroll(0, 0, remainingX - xOffset, remainingY - yOffset, 400);
-		post(this);
+		mStepper.prod();
 	}
 
 	public void smartMoveBackwards() {
@@ -274,7 +279,7 @@ public class ReaderView
 		}
 		mScrollerLastX = mScrollerLastY = 0;
 		mScroller.startScroll(0, 0, remainingX - xOffset, remainingY - yOffset, 400);
-		post(this);
+		mStepper.prod();
 	}
 
 	public void resetupChildren() {
@@ -289,19 +294,12 @@ public class ReaderView
 
 	public void refresh(boolean reflow) {
 		mReflow = reflow;
+		mReflowChanged = true;
+		mResetLayout = true;
 
 		mScale = 1.0f;
 		mXScroll = mYScroll = 0;
 
-		int numChildren = mChildViews.size();
-		for (int i = 0; i < numChildren; i++) {
-			View v = mChildViews.valueAt(i);
-			onNotInUse(v);
-			removeViewInLayout(v);
-		}
-		mChildViews.clear();
-		mViewCache.clear();
-
 		requestLayout();
 	}
 
@@ -337,7 +335,7 @@ public class ReaderView
 			mScrollerLastX = x;
 			mScrollerLastY = y;
 			requestLayout();
-			post(this);
+			mStepper.prod();
 		}
 		else if (!mUserInteracting) {
 			// End of an inertial scroll and the user is not interacting.
@@ -401,7 +399,7 @@ public class ReaderView
 			if(withinBoundsInDirectionOfTravel(bounds, velocityX, velocityY)
 					&& expandedBounds.contains(0, 0)) {
 				mScroller.fling(0, 0, (int)velocityX, (int)velocityY, bounds.left, bounds.right, bounds.top, bounds.bottom);
-				post(this);
+				mStepper.prod();
 			}
 		}
 
@@ -436,12 +434,9 @@ public class ReaderView
 		mScale = Math.min(Math.max(mScale * detector.getScaleFactor(), min_scale), max_scale);
 
 		if (mReflow) {
-			applyToChildren(new ViewMapper() {
-				@Override
-				void applyToView(View view) {
-					onScaleChild(view, mScale);
-				}
-			});
+			View v = mChildViews.get(mCurrent);
+			if (v != null)
+				onScaleChild(v, mScale);
 		} else {
 			float factor = mScale/previousScale;
 
@@ -472,6 +467,14 @@ public class ReaderView
 	}
 
 	public void onScaleEnd(ScaleGestureDetector detector) {
+		if (mReflow) {
+			applyToChildren(new ViewMapper() {
+				@Override
+				void applyToView(View view) {
+					onScaleChild(view, mScale);
+				}
+			});
+		}
 		mScaling = false;
 	}
 
@@ -482,10 +485,10 @@ public class ReaderView
 		if (!mScaling)
 			mGestureDetector.onTouchEvent(event);
 
-		if ((event.getAction() & event.ACTION_MASK) == MotionEvent.ACTION_DOWN) {
+		if ((event.getAction() & MotionEvent.ACTION_MASK) == MotionEvent.ACTION_DOWN) {
 			mUserInteracting = true;
 		}
-		if ((event.getAction() & event.ACTION_MASK) == MotionEvent.ACTION_UP) {
+		if ((event.getAction() & MotionEvent.ACTION_MASK) == MotionEvent.ACTION_UP) {
 			mScrollDisabled = false;
 			mUserInteracting = false;
 
@@ -537,7 +540,7 @@ public class ReaderView
 					postUnsettle(cv);
 					// post to invoke test for end of animation
 					// where we must set hq area for the new current view
-					post(this);
+					mStepper.prod();
 
 					onMoveOffChild(mCurrent);
 					mCurrent++;
@@ -548,7 +551,7 @@ public class ReaderView
 					postUnsettle(cv);
 					// post to invoke test for end of animation
 					// where we must set hq area for the new current view
-					post(this);
+					mStepper.prod();
 
 					onMoveOffChild(mCurrent);
 					mCurrent--;
@@ -585,8 +588,15 @@ public class ReaderView
 				removeViewInLayout(v);
 			}
 			mChildViews.clear();
+
+			// Don't reuse cached views if the adapter has changed
+			if (mReflowChanged) {
+				mReflowChanged = false;
+				mViewCache.clear();
+			}
+
 			// post to ensure generation of hq area
-			post(this);
+			mStepper.prod();
 		}
 
 		// Ensure current view is present
@@ -663,8 +673,7 @@ public class ReaderView
 	@Override
 	public void setAdapter(Adapter adapter) {
 		mAdapter = adapter;
-		mChildViews.clear();
-		removeAllViewsInLayout();
+
 		requestLayout();
 	}
 
@@ -772,7 +781,7 @@ public class ReaderView
 		if (corr.x != 0 || corr.y != 0) {
 			mScrollerLastX = mScrollerLastY = 0;
 			mScroller.startScroll(0, 0, corr.x, corr.y, 400);
-			post(this);
+			mStepper.prod();
 		}
 	}
 
