@@ -26,7 +26,7 @@ enum
 	PDFAPP_OUTLINE_LOAD_NOW = 2
 };
 
-static void pdfapp_showpage(pdfapp_t *app, int loadpage, int drawpage, int repaint, int transition);
+static void pdfapp_showpage(pdfapp_t *app, int loadpage, int drawpage, int repaint, int transition, int searching);
 static void pdfapp_updatepage(pdfapp_t *app);
 
 static void pdfapp_warn(pdfapp_t *app, const char *fmt, ...)
@@ -48,8 +48,8 @@ static void pdfapp_error(pdfapp_t *app, char *msg)
 char *pdfapp_version(pdfapp_t *app)
 {
 	return
-		"MuPDF 1.3\n"
-		"Copyright 2006-2013 Artifex Software, Inc.\n";
+		"MuPDF " FZ_VERSION "\n"
+		"Copyright 2006-2014 Artifex Software, Inc.\n";
 }
 
 char *pdfapp_usage(pdfapp_t *app)
@@ -192,6 +192,8 @@ void pdfapp_open_progressive(pdfapp_t *app, char *filename, int reload, int bps)
 	{
 		pdf_document *idoc;
 
+		fz_register_document_handlers(ctx);
+
 #ifdef HAVE_CURL
 		if (!strncmp(filename, "http://", 7))
 		{
@@ -247,7 +249,10 @@ void pdfapp_open_progressive(pdfapp_t *app, char *filename, int reload, int bps)
 		idoc = pdf_specifics(app->doc);
 
 		if (idoc)
+		{
+			pdf_enable_js(idoc);
 			pdf_set_doc_event_callback(idoc, event_cb, app);
+		}
 
 		if (fz_needs_password(app->doc))
 		{
@@ -328,7 +333,7 @@ void pdfapp_open_progressive(pdfapp_t *app, char *filename, int reload, int bps)
 		app->pany = 0;
 	}
 
-	pdfapp_showpage(app, 1, 1, 1, 0);
+	pdfapp_showpage(app, 1, 1, 1, 0, 0);
 }
 
 void pdfapp_close(pdfapp_t *app)
@@ -428,7 +433,7 @@ static int pdfapp_save(pdfapp_t *app)
 
 		if (gen_tmp_file(buf, PATH_MAX))
 		{
-			int written;
+			int written = 0;
 
 			fz_try(app->ctx)
 			{
@@ -438,7 +443,6 @@ static int pdfapp_save(pdfapp_t *app)
 			}
 			fz_catch(app->ctx)
 			{
-				written = 0;
 			}
 
 			if (written)
@@ -511,7 +515,7 @@ static void pdfapp_panview(pdfapp_t *app, int newx, int newy)
 	app->pany = newy;
 }
 
-static void pdfapp_loadpage(pdfapp_t *app)
+static void pdfapp_loadpage(pdfapp_t *app, int no_cache)
 {
 	fz_device *mdev = NULL;
 	int errored = 0;
@@ -560,6 +564,8 @@ static void pdfapp_loadpage(pdfapp_t *app)
 		/* Create display lists */
 		app->page_list = fz_new_display_list(app->ctx);
 		mdev = fz_new_list_device(app->ctx, app->page_list);
+		if (no_cache)
+			fz_enable_device_hints(mdev, FZ_NO_CACHE);
 		cookie.incomplete_ok = 1;
 		fz_run_page_contents(app->doc, app->page, mdev, &fz_identity, &cookie);
 		fz_free_device(mdev);
@@ -687,7 +693,7 @@ static void pdfapp_updatepage(pdfapp_t *app)
 		fz_free_device(idev);
 	}
 
-	pdfapp_showpage(app, 0, 0, 1, 0);
+	pdfapp_showpage(app, 0, 0, 1, 0, 0);
 }
 
 void pdfapp_reloadpage(pdfapp_t *app)
@@ -704,10 +710,10 @@ void pdfapp_reloadpage(pdfapp_t *app)
 		}
 		app->outline_deferred = 0;
 	}
-	pdfapp_showpage(app, 1, 1, 1, 0);
+	pdfapp_showpage(app, 1, 1, 1, 0, 0);
 }
 
-static void pdfapp_showpage(pdfapp_t *app, int loadpage, int drawpage, int repaint, int transition)
+static void pdfapp_showpage(pdfapp_t *app, int loadpage, int drawpage, int repaint, int transition, int searching)
 {
 	char buf[MAX_TITLE];
 	fz_device *idev;
@@ -732,7 +738,7 @@ static void pdfapp_showpage(pdfapp_t *app, int loadpage, int drawpage, int repai
 
 	if (loadpage)
 	{
-		pdfapp_loadpage(app);
+		pdfapp_loadpage(app, searching);
 
 		/* Zero search hit position */
 		app->hit_count = 0;
@@ -797,6 +803,10 @@ static void pdfapp_showpage(pdfapp_t *app, int loadpage, int drawpage, int repai
 		fz_transition *new_trans;
 		app->new_image = app->image;
 		app->image = NULL;
+		if (app->grayscale)
+			colorspace = fz_device_gray(app->ctx);
+		else
+			colorspace = app->colorspace;
 		app->image = fz_new_pixmap_with_bbox(app->ctx, colorspace, &ibounds);
 		app->duration = 0;
 		new_trans = fz_page_presentation(app->doc, app->page, &app->duration);
@@ -863,7 +873,7 @@ static void pdfapp_gotouri(pdfapp_t *app, char *uri)
 
 void pdfapp_gotopage(pdfapp_t *app, int number)
 {
-	app->isediting = 0;
+	app->issearching = 0;
 	winrepaint(app);
 
 	if (app->histlen + 1 == 256)
@@ -873,7 +883,7 @@ void pdfapp_gotopage(pdfapp_t *app, int number)
 	}
 	app->hist[app->histlen++] = app->pageno;
 	app->pageno = number + 1;
-	pdfapp_showpage(app, 1, 1, 1, 0);
+	pdfapp_showpage(app, 1, 1, 1, 0, 0);
 }
 
 void pdfapp_inverthit(pdfapp_t *app)
@@ -911,7 +921,7 @@ static void pdfapp_search_in_direction(pdfapp_t *app, enum panning *panto, int d
 		if (page != app->pageno)
 		{
 			app->pageno = page;
-			pdfapp_showpage(app, 1, 0, 0, 0);
+			pdfapp_showpage(app, 1, 0, 0, 0, 1);
 		}
 
 		app->hit_count = fz_search_text_page(app->ctx, app->page_text, app->search, app->hit_bbox, nelem(app->hit_bbox));
@@ -932,7 +942,7 @@ static void pdfapp_search_in_direction(pdfapp_t *app, enum panning *panto, int d
 	pdfapp_warn(app, "String '%s' not found.", app->search);
 
 	app->pageno = firstpage;
-	pdfapp_showpage(app, 1, 0, 0, 0);
+	pdfapp_showpage(app, 1, 0, 0, 0, 0);
 	wincursor(app, HAND);
 	winrepaint(app);
 }
@@ -954,7 +964,7 @@ void pdfapp_onkey(pdfapp_t *app, int c)
 	enum panning panto = PAN_TO_TOP;
 	int loadpage = 1;
 
-	if (app->isediting)
+	if (app->issearching)
 	{
 		int n = strlen(app->search);
 		if (c < ' ')
@@ -966,7 +976,7 @@ void pdfapp_onkey(pdfapp_t *app, int c)
 			}
 			if (c == '\n' || c == '\r')
 			{
-				app->isediting = 0;
+				app->issearching = 0;
 				if (n > 0)
 				{
 					winrepaintsearch(app);
@@ -977,7 +987,7 @@ void pdfapp_onkey(pdfapp_t *app, int c)
 							app->pageno = app->pagecount;
 						else
 							app->pageno--;
-						pdfapp_showpage(app, 1, 1, 0, 0);
+						pdfapp_showpage(app, 1, 1, 0, 0, 1);
 					}
 
 					pdfapp_onkey(app, 'n');
@@ -987,7 +997,7 @@ void pdfapp_onkey(pdfapp_t *app, int c)
 			}
 			if (c == '\033')
 			{
-				app->isediting = 0;
+				app->issearching = 0;
 				winrepaint(app);
 			}
 		}
@@ -1029,13 +1039,13 @@ void pdfapp_onkey(pdfapp_t *app, int c)
 		app->resolution *= ZOOMSTEP;
 		if (app->resolution > MAXRES)
 			app->resolution = MAXRES;
-		pdfapp_showpage(app, 0, 1, 1, 0);
+		pdfapp_showpage(app, 0, 1, 1, 0, 0);
 		break;
 	case '-':
 		app->resolution /= ZOOMSTEP;
 		if (app->resolution < MINRES)
 			app->resolution = MINRES;
-		pdfapp_showpage(app, 0, 1, 1, 0);
+		pdfapp_showpage(app, 0, 1, 1, 0, 0);
 		break;
 
 	case 'W':
@@ -1044,7 +1054,7 @@ void pdfapp_onkey(pdfapp_t *app, int c)
 			app->resolution = MAXRES;
 		else if (app->resolution < MINRES)
 			app->resolution = MINRES;
-		pdfapp_showpage(app, 0, 1, 1, 0);
+		pdfapp_showpage(app, 0, 1, 1, 0, 0);
 		break;
 	case 'H':
 		app->resolution *= (double) app->winh / (double) fz_pixmap_height(app->ctx, app->image);
@@ -1052,36 +1062,36 @@ void pdfapp_onkey(pdfapp_t *app, int c)
 			app->resolution = MAXRES;
 		else if (app->resolution < MINRES)
 			app->resolution = MINRES;
-		pdfapp_showpage(app, 0, 1, 1, 0);
+		pdfapp_showpage(app, 0, 1, 1, 0, 0);
 		break;
 
 	case 'L':
 		app->rotate -= 90;
-		pdfapp_showpage(app, 0, 1, 1, 0);
+		pdfapp_showpage(app, 0, 1, 1, 0, 0);
 		break;
 	case 'R':
 		app->rotate += 90;
-		pdfapp_showpage(app, 0, 1, 1, 0);
+		pdfapp_showpage(app, 0, 1, 1, 0, 0);
 		break;
 
 	case 'c':
 		app->grayscale ^= 1;
-		pdfapp_showpage(app, 0, 1, 1, 0);
+		pdfapp_showpage(app, 0, 1, 1, 0, 0);
 		break;
 
 	case 'i':
 		app->invert ^= 1;
-		pdfapp_showpage(app, 0, 1, 1, 0);
+		pdfapp_showpage(app, 0, 1, 1, 0, 0);
 		break;
 
 #ifndef NDEBUG
 	case 'a':
 		app->rotate -= 15;
-		pdfapp_showpage(app, 0, 1, 1, 0);
+		pdfapp_showpage(app, 0, 1, 1, 0, 0);
 		break;
 	case 's':
 		app->rotate += 15;
-		pdfapp_showpage(app, 0, 1, 1, 0);
+		pdfapp_showpage(app, 0, 1, 1, 0, 0);
 		break;
 #endif
 
@@ -1103,27 +1113,27 @@ void pdfapp_onkey(pdfapp_t *app, int c)
 		}
 		app->shrinkwrap = 1;
 		app->panx = app->pany = 0;
-		pdfapp_showpage(app, 0, 0, 1, 0);
+		pdfapp_showpage(app, 0, 0, 1, 0, 0);
 		break;
 
 	case 'h':
 		app->panx += fz_pixmap_width(app->ctx, app->image) / 10;
-		pdfapp_showpage(app, 0, 0, 1, 0);
+		pdfapp_showpage(app, 0, 0, 1, 0, 0);
 		break;
 
 	case 'j':
 		app->pany -= fz_pixmap_height(app->ctx, app->image) / 10;
-		pdfapp_showpage(app, 0, 0, 1, 0);
+		pdfapp_showpage(app, 0, 0, 1, 0, 0);
 		break;
 
 	case 'k':
 		app->pany += fz_pixmap_height(app->ctx, app->image) / 10;
-		pdfapp_showpage(app, 0, 0, 1, 0);
+		pdfapp_showpage(app, 0, 0, 1, 0, 0);
 		break;
 
 	case 'l':
 		app->panx -= fz_pixmap_width(app->ctx, app->image) / 10;
-		pdfapp_showpage(app, 0, 0, 1, 0);
+		pdfapp_showpage(app, 0, 0, 1, 0, 0);
 		break;
 
 	/*
@@ -1247,7 +1257,7 @@ void pdfapp_onkey(pdfapp_t *app, int c)
 	 */
 
 	case '?':
-		app->isediting = 1;
+		app->issearching = 1;
 		app->searchdir = -1;
 		app->search[0] = 0;
 		app->hit_count = 0;
@@ -1256,7 +1266,7 @@ void pdfapp_onkey(pdfapp_t *app, int c)
 		break;
 
 	case '/':
-		app->isediting = 1;
+		app->issearching = 1;
 		app->searchdir = 1;
 		app->search[0] = 0;
 		app->hit_count = 0;
@@ -1303,7 +1313,7 @@ void pdfapp_onkey(pdfapp_t *app, int c)
 		case DONT_PAN:
 			break;
 		}
-		pdfapp_showpage(app, loadpage, 1, 1, 1);
+		pdfapp_showpage(app, loadpage, 1, 1, 1, 0);
 	}
 }
 
@@ -1412,7 +1422,7 @@ void pdfapp_onmouse(pdfapp_t *app, int x, int y, int btn, int modifiers, int sta
 						char ebuf[256];
 
 						ebuf[0] = 0;
-						if (pdf_check_signature(ctx, idoc, widget, app->docpath, ebuf, sizeof(ebuf)))
+						if (pdf_check_signature(idoc, widget, app->docpath, ebuf, sizeof(ebuf)))
 						{
 							winwarn(app, "Signature is valid");
 						}
@@ -1503,7 +1513,7 @@ void pdfapp_onmouse(pdfapp_t *app, int x, int y, int btn, int modifiers, int sta
 					app->resolution = MAXRES;
 				if (app->resolution < MINRES)
 					app->resolution = MINRES;
-				pdfapp_showpage(app, 0, 1, 1, 0);
+				pdfapp_showpage(app, 0, 1, 1, 0, 0);
 			}
 			else
 			{
@@ -1515,6 +1525,15 @@ void pdfapp_onmouse(pdfapp_t *app, int x, int y, int btn, int modifiers, int sta
 				pdfapp_panview(app, app->panx + xstep, app->pany + ystep);
 			}
 		}
+		if (btn == 6 || btn == 7) /* scroll wheel (horizontal) */
+		{
+			/* scroll left/right or up/down if shift is pressed */
+			int dir = btn == 6 ? 1 : -1;
+			int isx = (modifiers & (1<<0));
+			int xstep = !isx ? 20 * dir : 0;
+			int ystep = isx ? 20 * dir : 0;
+			pdfapp_panview(app, app->panx + xstep, app->pany + ystep);
+		}
 	}
 
 	else if (state == -1)
@@ -1560,7 +1579,7 @@ void pdfapp_onmouse(pdfapp_t *app, int x, int y, int btn, int modifiers, int sta
 					if( app->pageno > 1 )
 					{
 						app->pageno--;
-						pdfapp_showpage(app, 1, 1, 1, 0);
+						pdfapp_showpage(app, 1, 1, 1, 0, 0);
 						newy = -fz_pixmap_height(app->ctx, app->image);
 					}
 					app->beyondy = 0;
@@ -1570,7 +1589,7 @@ void pdfapp_onmouse(pdfapp_t *app, int x, int y, int btn, int modifiers, int sta
 					if( app->pageno < app->pagecount )
 					{
 						app->pageno++;
-						pdfapp_showpage(app, 1, 1, 1, 0);
+						pdfapp_showpage(app, 1, 1, 1, 0, 0);
 						newy = 0;
 					}
 					app->beyondy = 0;
