@@ -1,32 +1,35 @@
 #include "mupdf/fitz.h"
 
 void
-fz_eval_function(fz_context *ctx, fz_function *func, float *in_, int inlen, float *out_, int outlen)
+fz_eval_function(fz_context *ctx, fz_function *func, const float *in, int inlen, float *out, int outlen)
 {
 	float fakein[FZ_FN_MAXM];
 	float fakeout[FZ_FN_MAXN];
-	float *in = in_;
-	float *out = out_;
+	int i;
 
 	if (inlen < func->m)
 	{
+		for (i = 0; i < func->m; ++i)
+			fakein[i] = in[i];
+		for (; i < inlen; ++i)
+			fakein[i] = 0;
 		in = fakein;
-		memset(in, 0, sizeof(float) * func->m);
-		memcpy(in, in_, sizeof(float) * inlen);
 	}
 
 	if (outlen < func->n)
 	{
-		out = fakeout;
-		memset(out, 0, sizeof(float) * func->n);
+		func->evaluate(ctx, func, in, fakeout);
+		for (i = 0; i < func->n; ++i)
+			out[i] = fakeout[i];
+		for (; i < outlen; ++i)
+			out[i] = 0;
 	}
 	else
-		memset(out, 0, sizeof(float) * outlen);
-
-	func->evaluate(ctx, func, in, out);
-
-	if (outlen < func->n)
-		memcpy(out_, out, sizeof(float) * outlen);
+	{
+		func->evaluate(ctx, func, in, out);
+		for (i = func->n; i < outlen; ++i)
+			out[i] = 0;
+	}
 }
 
 fz_function *
