@@ -1,5 +1,11 @@
 #include "mupdf/pdf.h"
 
+#include <ft2build.h>
+#include FT_FREETYPE_H
+#include FT_ADVANCES_H
+
+#define ALLOWED_TEXT_POS_ERROR (0.001f)
+
 typedef struct pdf_device_s pdf_device;
 
 typedef struct gstate_s gstate;
@@ -57,7 +63,6 @@ typedef struct group_entry_s group_entry;
 
 struct group_entry_s
 {
-	int blendmode;
 	int alpha;
 	int isolated;
 	int knockout;
@@ -361,30 +366,30 @@ pdf_dev_path(pdf_device *pdev, fz_path *path)
 	fz_context *ctx = pdev->ctx;
 	gstate *gs = CURRENT_GSTATE(pdev);
 	float x, y;
-	int i = 0;
-	while (i < path->len)
+	int i = 0, k = 0;
+	while (i < path->cmd_len)
 	{
-		switch (path->items[i++].k)
+		switch (path->cmds[i++])
 		{
 		case FZ_MOVETO:
-			x = path->items[i++].v;
-			y = path->items[i++].v;
+			x = path->coords[k++];
+			y = path->coords[k++];
 			fz_buffer_printf(ctx, gs->buf, "%f %f m\n", x, y);
 			break;
 		case FZ_LINETO:
-			x = path->items[i++].v;
-			y = path->items[i++].v;
+			x = path->coords[k++];
+			y = path->coords[k++];
 			fz_buffer_printf(ctx, gs->buf, "%f %f l\n", x, y);
 			break;
 		case FZ_CURVETO:
-			x = path->items[i++].v;
-			y = path->items[i++].v;
+			x = path->coords[k++];
+			y = path->coords[k++];
 			fz_buffer_printf(ctx, gs->buf, "%f %f ", x, y);
-			x = path->items[i++].v;
-			y = path->items[i++].v;
+			x = path->coords[k++];
+			y = path->coords[k++];
 			fz_buffer_printf(ctx, gs->buf, "%f %f ", x, y);
-			x = path->items[i++].v;
-			y = path->items[i++].v;
+			x = path->coords[k++];
+			y = path->coords[k++];
 			fz_buffer_printf(ctx, gs->buf, "%f %f c\n", x, y);
 			break;
 		case FZ_CLOSE_PATH:
@@ -543,6 +548,9 @@ pdf_dev_font(pdf_device *pdev, fz_font *font, float size)
 	if (gs->font >= 0 && pdev->fonts[gs->font].font == font)
 		return;
 
+	if (font->ft_buffer != NULL || font->ft_substitute)
+		fz_throw(pdev->ctx, FZ_ERROR_GENERIC, "pdf device supports only base 14 fonts currently");
+
 	/* Have we sent such a font before? */
 	for (i = 0; i < pdev->num_fonts; i++)
 		if (pdev->fonts[i].font == font)
@@ -569,11 +577,11 @@ pdf_dev_font(pdf_device *pdev, fz_font *font, float size)
 		o = pdf_new_dict(doc, 3);
 		fz_try(ctx)
 		{
-			/* BIG FIXME: Get someone who understands fonts to fill this bit in. */
 			char text[32];
 			pdf_dict_puts_drop(o, "Type", pdf_new_name(doc, "Font"));
 			pdf_dict_puts_drop(o, "Subtype", pdf_new_name(doc, "Type1"));
-			pdf_dict_puts_drop(o, "BaseFont", pdf_new_name(doc, "Helvetica"));
+			pdf_dict_puts_drop(o, "BaseFont", pdf_new_name(doc, font->name));
+			pdf_dict_puts_drop(o, "Encoding", pdf_new_name(doc, "WinAnsiEncoding"));
 			ref = pdf_new_ref(doc, o);
 			snprintf(text, sizeof(text), "Font/F%d", i);
 			pdf_dict_putp(pdev->resources, text, ref);
@@ -650,15 +658,15 @@ pdf_dev_pop(pdf_device *pdev)
 }
 
 static void
-pdf_dev_text(pdf_device *pdev, fz_text *text)
+pdf_dev_text(pdf_device *pdev, fz_text *text, float size)
 {
+	int mask = FT_LOAD_NO_SCALE | FT_LOAD_IGNORE_TRANSFORM;
 	int i;
 	fz_matrix trm;
 	fz_matrix inverse;
 	gstate *gs = CURRENT_GSTATE(pdev);
 	fz_matrix trunc_trm;
 
-	/* BIG FIXME: Get someone who understands fonts to fill this bit in. */
 	trm = gs->tm;
 	trunc_trm.a = trm.a;
 	trunc_trm.b = trm.b;
@@ -668,10 +676,14 @@ pdf_dev_text(pdf_device *pdev, fz_text *text)
 	trunc_trm.f = 0;
 	fz_invert_matrix(&inverse, &trunc_trm);
 
-	for (i=0; i < text->len; i++)
+	i = 0;
+	while (i < text->len)
 	{
 		fz_text_item *it = &text->items[i];
 		fz_point delta;
+		float x;
+		int j;
+
 		delta.x = it->x - trm.e;
 		delta.y = it->y - trm.f;
 		fz_transform_point(&delta, &inverse);
@@ -681,10 +693,32 @@ pdf_dev_text(pdf_device *pdev, fz_text *text)
 			trm.e = it->x;
 			trm.f = it->y;
 		}
-		fz_buffer_printf(pdev->ctx, gs->buf, "<%02x> Tj\n", it->ucs);
-		/* FIXME: Advance the text position - doesn't matter at the
-		 * moment as we absolutely position each glyph, but we should
-		 * use more efficient text outputting where possible. */
+
+		j = i+1;
+		if (text->font->ft_face)
+		{
+			/* Find prefix of text for which the advance of each character accounts
+			 * for the position offset */
+			x = it->x;
+			while (j < text->len)
+			{
+				FT_Fixed adv;
+				FT_Get_Advance(text->font->ft_face, text->items[j-1].gid, mask, &adv);
+				x += (float)adv * size /((FT_Face)text->font->ft_face)->units_per_EM;
+				if (fabs(x - text->items[j].x) > ALLOWED_TEXT_POS_ERROR || fabs(it->y - text->items[j].y) > ALLOWED_TEXT_POS_ERROR)
+					break;
+				j++;
+			}
+		}
+
+		fz_buffer_printf(pdev->ctx, gs->buf, "<");
+		for (/* i from its current value */; i < j; i++)
+		{
+			/* FIXME: should use it->gid, rather than it->ucs, and convert
+			* to the correct encoding */
+			fz_buffer_printf(pdev->ctx, gs->buf, "%02x", text->items[i].ucs);
+		}
+		fz_buffer_printf(pdev->ctx, gs->buf, "> Tj\n");
 	}
 	gs->tm.e = trm.e;
 	gs->tm.f = trm.f;
@@ -732,7 +766,7 @@ pdf_dev_end_text(pdf_device *pdev)
 }
 
 static int
-pdf_dev_new_form(pdf_obj **form_ref, pdf_device *pdev, const fz_rect *bbox, int isolated, int knockout, int blendmode, float alpha, fz_colorspace *colorspace)
+pdf_dev_new_form(pdf_obj **form_ref, pdf_device *pdev, const fz_rect *bbox, int isolated, int knockout, float alpha, fz_colorspace *colorspace)
 {
 	fz_context *ctx = pdev->ctx;
 	pdf_document *doc = pdev->doc;
@@ -747,7 +781,7 @@ pdf_dev_new_form(pdf_obj **form_ref, pdf_device *pdev, const fz_rect *bbox, int
 	for(num = 0; num < pdev->num_groups; num++)
 	{
 		group_entry *g = &pdev->groups[num];
-		if (g->isolated == isolated && g->knockout == knockout && g->blendmode == blendmode && g->alpha == alpha && g->colorspace == colorspace)
+		if (g->isolated == isolated && g->knockout == knockout && g->alpha == alpha && g->colorspace == colorspace)
 		{
 			group_ref = pdev->groups[num].ref;
 			break;
@@ -768,7 +802,6 @@ pdf_dev_new_form(pdf_obj **form_ref, pdf_device *pdev, const fz_rect *bbox, int
 		pdev->num_groups++;
 		pdev->groups[num].isolated = isolated;
 		pdev->groups[num].knockout = knockout;
-		pdev->groups[num].blendmode = blendmode;
 		pdev->groups[num].alpha = alpha;
 		pdev->groups[num].colorspace = fz_keep_colorspace(ctx, colorspace);
 		pdev->groups[num].ref = NULL;
@@ -779,8 +812,6 @@ pdf_dev_new_form(pdf_obj **form_ref, pdf_device *pdev, const fz_rect *bbox, int
 			pdf_dict_puts_drop(group, "S", pdf_new_name(doc, "Transparency"));
 			pdf_dict_puts_drop(group, "K", pdf_new_bool(doc, knockout));
 			pdf_dict_puts_drop(group, "I", pdf_new_bool(doc, isolated));
-			pdf_dict_puts_drop(group, "K", pdf_new_bool(doc, knockout));
-			pdf_dict_puts_drop(group, "BM", pdf_new_name(doc, fz_blendmode_name(blendmode)));
 			if (!colorspace)
 			{}
 			else if (colorspace->n == 1)
@@ -899,13 +930,17 @@ pdf_dev_fill_text(fz_device *dev, fz_text *text, const fz_matrix *ctm,
 	fz_colorspace *colorspace, float *color, float alpha)
 {
 	pdf_device *pdev = dev->user;
+	fz_matrix trm = text->trm;
+	float size = fz_matrix_expansion(&trm);
 
-	pdf_dev_begin_text(pdev, &text->trm, 0);
-	pdf_dev_font(pdev, text->font, 1);
+	fz_pre_scale(&trm, 1/size, 1/size);
+
+	pdf_dev_begin_text(pdev, &trm, 0);
+	pdf_dev_font(pdev, text->font, size);
 	pdf_dev_ctm(pdev, ctm);
 	pdf_dev_alpha(pdev, alpha, 0);
 	pdf_dev_color(pdev, colorspace, color, 0);
-	pdf_dev_text(pdev, text);
+	pdf_dev_text(pdev, text, size);
 }
 
 static void
@@ -913,46 +948,62 @@ pdf_dev_stroke_text(fz_device *dev, fz_text *text, fz_stroke_state *stroke, cons
 	fz_colorspace *colorspace, float *color, float alpha)
 {
 	pdf_device *pdev = dev->user;
+	fz_matrix trm = text->trm;
+	float size = fz_matrix_expansion(&trm);
+
+	fz_pre_scale(&trm, 1/size, 1/size);
 
 	pdf_dev_begin_text(pdev, &text->trm, 1);
 	pdf_dev_font(pdev, text->font, 1);
 	pdf_dev_ctm(pdev, ctm);
 	pdf_dev_alpha(pdev, alpha, 1);
 	pdf_dev_color(pdev, colorspace, color, 1);
-	pdf_dev_text(pdev, text);
+	pdf_dev_text(pdev, text, size);
 }
 
 static void
 pdf_dev_clip_text(fz_device *dev, fz_text *text, const fz_matrix *ctm, int accumulate)
 {
 	pdf_device *pdev = dev->user;
+	fz_matrix trm = text->trm;
+	float size = fz_matrix_expansion(&trm);
+
+	fz_pre_scale(&trm, 1/size, 1/size);
 
 	pdf_dev_begin_text(pdev, &text->trm, 0);
 	pdf_dev_ctm(pdev, ctm);
 	pdf_dev_font(pdev, text->font, 7);
-	pdf_dev_text(pdev, text);
+	pdf_dev_text(pdev, text, size);
 }
 
 static void
 pdf_dev_clip_stroke_text(fz_device *dev, fz_text *text, fz_stroke_state *stroke, const fz_matrix *ctm)
 {
 	pdf_device *pdev = dev->user;
+	fz_matrix trm = text->trm;
+	float size = fz_matrix_expansion(&trm);
+
+	fz_pre_scale(&trm, 1/size, 1/size);
 
 	pdf_dev_begin_text(pdev, &text->trm, 0);
 	pdf_dev_font(pdev, text->font, 5);
 	pdf_dev_ctm(pdev, ctm);
-	pdf_dev_text(pdev, text);
+	pdf_dev_text(pdev, text, size);
 }
 
 static void
 pdf_dev_ignore_text(fz_device *dev, fz_text *text, const fz_matrix *ctm)
 {
 	pdf_device *pdev = dev->user;
+	fz_matrix trm = text->trm;
+	float size = fz_matrix_expansion(&trm);
+
+	fz_pre_scale(&trm, 1/size, 1/size);
 
 	pdf_dev_begin_text(pdev, &text->trm, 0);
 	pdf_dev_ctm(pdev, ctm);
 	pdf_dev_font(pdev, text->font, 3);
-	pdf_dev_text(pdev, text);
+	pdf_dev_text(pdev, text, size);
 }
 
 static void
@@ -1044,7 +1095,7 @@ pdf_dev_begin_mask(fz_device *dev, const fz_rect *bbox, int luminosity, fz_color
 	pdf_dev_end_text(pdev);
 
 	/* Make a new form to contain the contents of the softmask */
-	pdf_dev_new_form(&form_ref, pdev, bbox, 0, 0, 0, 1, colorspace);
+	pdf_dev_new_form(&form_ref, pdev, bbox, 0, 0, 1, colorspace);
 
 	fz_try(ctx)
 	{
@@ -1115,17 +1166,34 @@ pdf_dev_begin_group(fz_device *dev, const fz_rect *bbox, int isolated, int knock
 {
 	pdf_device *pdev = (pdf_device *)dev->user;
 	fz_context *ctx = pdev->ctx;
+	pdf_document *doc = pdev->doc;
 	int num;
 	pdf_obj *form_ref;
 	gstate *gs;
 
 	pdf_dev_end_text(pdev);
 
-	num = pdf_dev_new_form(&form_ref, pdev, bbox, isolated, knockout, blendmode, alpha, NULL);
+	num = pdf_dev_new_form(&form_ref, pdev, bbox, isolated, knockout, alpha, NULL);
+
+	/* Do we have an appropriate blending extgstate already? */
+	{
+		char text[32];
+		pdf_obj *obj;
+		snprintf(text, sizeof(text), "ExtGState/BlendMode%d", blendmode);
+		obj = pdf_dict_getp(pdev->resources, text);
+		if (obj == NULL)
+		{
+			/* No, better make one */
+			obj = pdf_new_dict(pdev->doc, 2);
+			pdf_dict_puts_drop(obj, "Type", pdf_new_name(doc, "ExtGState"));
+			pdf_dict_puts_drop(obj, "BM", pdf_new_name(doc, fz_blendmode_name(blendmode)));
+			pdf_dict_putp_drop(pdev->resources, text, obj);
+		}
+	}
 
 	/* Add the call to this group */
 	gs = CURRENT_GSTATE(pdev);
-	fz_buffer_printf(dev->ctx, gs->buf, "/Fm%d Do\n", num);
+	fz_buffer_printf(dev->ctx, gs->buf, "/BlendMode%d gs /Fm%d Do\n", blendmode, num);
 
 	/* Now, everything we get until the end of group needs to go into a
 	 * new buffer, which will be the stream contents for the form. */
@@ -1209,6 +1277,14 @@ pdf_dev_free_user(fz_device *dev)
 	fz_free(ctx, pdev);
 }
 
+static void
+pdf_dev_rebind(fz_device *dev)
+{
+	pdf_device *pdev = dev->user;
+
+	fz_rebind_document((fz_document *)pdev->doc, dev->ctx);
+}
+
 fz_device *pdf_new_pdf_device(pdf_document *doc, pdf_obj *contents, pdf_obj *resources, const fz_matrix *ctm)
 {
 	fz_context *ctx = doc->ctx;
@@ -1245,6 +1321,7 @@ fz_device *pdf_new_pdf_device(pdf_document *doc, pdf_obj *contents, pdf_obj *res
 		fz_rethrow(ctx);
 	}
 
+	dev->rebind = pdf_dev_rebind;
 	dev->free_user = pdf_dev_free_user;
 
 	dev->fill_path = pdf_dev_fill_path;
@@ -1278,14 +1355,34 @@ fz_device *pdf_new_pdf_device(pdf_document *doc, pdf_obj *contents, pdf_obj *res
 
 fz_device *pdf_page_write(pdf_document *doc, pdf_page *page)
 {
+	fz_context *ctx = doc->ctx;
 	pdf_obj *resources = pdf_dict_gets(page->me, "Resources");
 	fz_matrix ctm;
 	fz_pre_translate(fz_scale(&ctm, 1, -1), 0, page->mediabox.y0-page->mediabox.y1);
+
 	if (resources == NULL)
 	{
 		resources = pdf_new_dict(doc, 0);
 		pdf_dict_puts_drop(page->me, "Resources", resources);
 	}
 
+	if (page->contents == NULL)
+	{
+		pdf_obj *obj = pdf_new_dict(doc, 0);
+		fz_try(ctx)
+		{
+			page->contents = pdf_new_ref(doc, obj);
+			pdf_dict_puts(page->me, "Contents", page->contents);
+		}
+		fz_always(ctx)
+		{
+			pdf_drop_obj(obj);
+		}
+		fz_catch(ctx)
+		{
+			fz_rethrow(ctx);
+		}
+	}
+
 	return pdf_new_pdf_device(doc, page->contents, resources, &ctm);
 }
